from __future__ import annotations

import gzip
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO

import pytest

if TYPE_CHECKING:
    from collections.abc import Iterator


def absolute_path(filename: str) -> Path:
    return Path(__file__).parent / filename


def open_file(name: str, mode: str = "rb") -> Iterator[BinaryIO]:
    with absolute_path(name).open(mode) as fh:
        yield fh


def open_file_gz(name: str, mode: str = "rb") -> Iterator[BinaryIO]:
    with gzip.GzipFile(absolute_path(name), mode) as fh:
        yield fh


@pytest.fixture
def gzip_sqfs() -> Iterator[BinaryIO]:
    yield from open_file("_data/gzip.sqfs")


@pytest.fixture
def gzip_opts_sqfs() -> Iterator[BinaryIO]:
    yield from open_file("_data/gzip-opts.sqfs")


@pytest.fixture
def lz4_sqfs() -> Iterator[BinaryIO]:
    yield from open_file("_data/lz4.sqfs")


@pytest.fixture
def lzma_sqfs() -> Iterator[BinaryIO]:
    yield from open_file("_data/lzma.sqfs")


@pytest.fixture
def lzo_sqfs() -> Iterator[BinaryIO]:
    yield from open_file("_data/lzo.sqfs")


@pytest.fixture
def xz_sqfs() -> Iterator[BinaryIO]:
    yield from open_file("_data/xz.sqfs")


@pytest.fixture
def zstd_sqfs() -> Iterator[BinaryIO]:
    yield from open_file("_data/zstd.sqfs")



from __future__ import annotations

from io import BytesIO

import pytest

from dissect.squashfs.c_squashfs import c_squashfs
from dissect.squashfs.exceptions import (
    NotADirectoryError,
    NotAFileError,
    NotASymlinkError,
)
from dissect.squashfs.squashfs import SquashFS


def _verify_filesystem(sqfs: SquashFS) -> None:
    assert sqfs.root.is_dir()
    assert sorted(sqfs.root.listdir().keys()) == [
        "dir",
        "file-with-xattr",
        "large-file",
        "small-file",
        "symlink-1",
        "symlink-2",
        "symlink-with-xattr",
    ]

    f1 = sqfs.get("small-file")
    assert f1.is_file()
    assert f1.size == 9
    assert str(f1.mtime) == "2022-12-05 18:53:05+00:00"
    assert f1.open().read() == b"contents\n"
    f2 = sqfs.get("large-file")
    assert f2.is_file()
    assert f2.size == 4177920
    assert f2.open().read() == b"".join([bytes([i] * 4096) for i in range(255)]) * 4

    s1 = sqfs.get("symlink-1")
    assert s1.is_symlink()
    assert s1.link == "small-file"
    s2 = sqfs.get("symlink-2")
    assert s2.is_symlink()
    assert s2.link == "dir/file_69"

    d1 = sqfs.get("dir")
    assert d1.is_dir()
    assert d1.size == 1507
    assert len(list(d1.iterdir())) == 100

    assert sqfs.get("dir/file_69").inode_number == s2.link_inode.inode_number

    with pytest.raises(NotAFileError):
        sqfs.get("dir").open()

    with pytest.raises(NotADirectoryError):
        sqfs.get("small-file").listdir()

    with pytest.raises(NotASymlinkError):
        assert sqfs.get("large-file").link


@pytest.mark.parametrize(
    ("sqfs", "compression_id"),
    [
        ("gzip_sqfs", c_squashfs.ZLIB_COMPRESSION),
        ("gzip_opts_sqfs", c_squashfs.ZLIB_COMPRESSION),
        ("lz4_sqfs", c_squashfs.LZ4_COMPRESSION),
        # Whether LZMA works seems to depend on if your machine has been blessed by the compression gods
        # Include the test data for reference material, but don't test against it
        # ("lzma_sqfs", c_squashfs.LZMA_COMPRESSION),
        ("lzo_sqfs", c_squashfs.LZO_COMPRESSION),
        ("xz_sqfs", c_squashfs.XZ_COMPRESSION),
        ("zstd_sqfs", c_squashfs.ZSTD_COMPRESSION),
    ],
)
def test_squashfs(sqfs: str, compression_id: int, request: pytest.FixtureRequest) -> None:
    sqfs = SquashFS(request.getfixturevalue(sqfs))
    assert sqfs.sb.compression == compression_id
    _verify_filesystem(sqfs)


def test_squashfs_unsupported_bigendian_3_0() -> None:
    """Test if we correctly detect a SquashFS 3.0 big-endian filesystem."""

    # Random SquashFS version 3.0 big-endian header found on binwalk GitHub issue tracker.
    buf = bytes.fromhex(
        "73717368000000b60000002042010d364200bc844200339440016800000300006b0400104003004ae64f330000000000"
        "001623000100000000001177ff8e0800000000000e303100000000000e3025000000000000000000000000000e234c00"
    )

    sb = c_squashfs.squashfs_super_block(buf)
    assert sb.s_magic == c_squashfs.SQUASHFS_MAGIC_SWAP
    assert sb.s_major == 0x300

    with pytest.raises(NotImplementedError, match="Unsupported squashfs pre-4.0 big-endian filesystem"):
        SquashFS(BytesIO(buf))



from __future__ import annotations

import stat

from dissect.cstruct import cstruct

squashfs_def = """
#define SQUASHFS_MAGIC          0x73717368
#define SQUASHFS_MAGIC_SWAP     0x68737173

/* size of metadata (inode and directory) blocks */
#define SQUASHFS_METADATA_SIZE  8192

#define SQUASHFS_INVALID        0xffffffffffff
#define SQUASHFS_INVALID_FRAG   0xffffffff
#define SQUASHFS_INVALID_XATTR  0xffffffff
#define SQUASHFS_INVALID_BLK    -1
#define SQUASHFS_USED_BLK       -2

/* Filesystem flags */
#define SQUASHFS_NOI            0
#define SQUASHFS_NOD            1
#define SQUASHFS_CHECK          2
#define SQUASHFS_NOF            3
#define SQUASHFS_NO_FRAG        4
#define SQUASHFS_ALWAYS_FRAG    5
#define SQUASHFS_DUPLICATE      6
#define SQUASHFS_EXPORT         7
#define SQUASHFS_NOX            8
#define SQUASHFS_NO_XATTR       9
#define SQUASHFS_COMP_OPT       10
#define SQUASHFS_NOID           11

/* Max number of types and file types */
#define SQUASHFS_DIR_TYPE       1
#define SQUASHFS_FILE_TYPE      2
#define SQUASHFS_SYMLINK_TYPE   3
#define SQUASHFS_BLKDEV_TYPE    4
#define SQUASHFS_CHRDEV_TYPE    5
#define SQUASHFS_FIFO_TYPE      6
#define SQUASHFS_SOCKET_TYPE    7
#define SQUASHFS_LDIR_TYPE      8
#define SQUASHFS_LREG_TYPE      9
#define SQUASHFS_LSYMLINK_TYPE  10
#define SQUASHFS_LBLKDEV_TYPE   11
#define SQUASHFS_LCHRDEV_TYPE   12
#define SQUASHFS_LFIFO_TYPE     13
#define SQUASHFS_LSOCKET_TYPE   14

/* Xattr types */
#define SQUASHFS_XATTR_USER     0
#define SQUASHFS_XATTR_TRUSTED  1
#define SQUASHFS_XATTR_SECURITY 2
#define SQUASHFS_XATTR_VALUE_OOL    256
#define SQUASHFS_XATTR_PREFIX_MASK  0xff

/* Flag whether block is compressed or uncompressed, bit is set if block is uncompressed */
#define SQUASHFS_COMPRESSED_BIT         (1 << 15)
#define SQUASHFS_COMPRESSED_BIT_BLOCK   (1 << 24)

typedef long long squashfs_block;
typedef long long squashfs_inode;

#define ZLIB_COMPRESSION    1
#define LZMA_COMPRESSION    2
#define LZO_COMPRESSION     3
#define XZ_COMPRESSION      4
#define LZ4_COMPRESSION     5
#define ZSTD_COMPRESSION    6

struct squashfs_super_block {
    unsigned int        s_magic;
    unsigned int        inodes;
    unsigned int        mkfs_time;
    unsigned int        block_size;
    unsigned int        fragments;
    unsigned short      compression;
    unsigned short      block_log;
    unsigned short      flags;
    unsigned short      no_ids;
    unsigned short      s_major;
    unsigned short      s_minor;
    squashfs_inode      root_inode;
    long long           bytes_used;
    long long           id_table_start;
    long long           xattr_id_table_start;
    long long           inode_table_start;
    long long           directory_table_start;
    long long           fragment_table_start;
    long long           lookup_table_start;
};

struct squashfs_super_block_3 {
    unsigned int        s_magic;
    unsigned int        inodes;
    unsigned int        bytes_used_2;
    unsigned int        uid_start_2;
    unsigned int        guid_start_2;
    unsigned int        inode_table_start_2;
    unsigned int        directory_table_start_2;
    unsigned int        s_major : 16;
    unsigned int        s_minor : 16;
    unsigned int        block_size_1 : 16;
    unsigned int        block_log : 16;
    unsigned int        flags : 8;
    unsigned int        no_uids : 8;
    unsigned int        no_guids : 8;
    int                 mkfs_time;
    squashfs_inode      root_inode;
    unsigned int        block_size;
    unsigned int        fragments;
    unsigned int        fragment_table_start_2;
    long long           bytes_used;
    long long           uid_start;
    long long           guid_start;
    long long           inode_table_start;
    long long           directory_table_start;
    long long           fragment_table_start;
    long long           lookup_table_start;
};

struct squashfs_dir_index {
    unsigned int        index;
    unsigned int        start_block;
    unsigned int        size;
    unsigned char       name[0];
};

struct squashfs_base_inode_header {
    unsigned short      inode_type;
    unsigned short      mode;
    unsigned short      uid;
    unsigned short      guid;
    unsigned int        mtime;
    unsigned int        inode_number;
};

struct squashfs_ipc_inode_header {
    unsigned short      inode_type;
    unsigned short      mode;
    unsigned short      uid;
    unsigned short      guid;
    unsigned int        mtime;
    unsigned int        inode_number;
    unsigned int        nlink;
};

struct squashfs_lipc_inode_header {
    unsigned short      inode_type;
    unsigned short      mode;
    unsigned short      uid;
    unsigned short      guid;
    unsigned int        mtime;
    unsigned int        inode_number;
    unsigned int        nlink;
    unsigned int        xattr;
};

struct squashfs_dev_inode_header {
    unsigned short      inode_type;
    unsigned short      mode;
    unsigned short      uid;
    unsigned short      guid;
    unsigned int        mtime;
    unsigned int        inode_number;
    unsigned int        nlink;
    unsigned int        rdev;
};

struct squashfs_ldev_inode_header {
    unsigned short      inode_type;
    unsigned short      mode;
    unsigned short      uid;
    unsigned short      guid;
    unsigned int        mtime;
    unsigned int        inode_number;
    unsigned int        nlink;
    unsigned int        rdev;
    unsigned int        xattr;
};

struct squashfs_symlink_inode_header {
    unsigned short      inode_type;
    unsigned short      mode;
    unsigned short      uid;
    unsigned short      guid;
    unsigned int        mtime;
    unsigned int        inode_number;
    unsigned int        nlink;
    unsigned int        symlink_size;
    char                symlink[0];
};

struct squashfs_reg_inode_header {
    unsigned short      inode_type;
    unsigned short      mode;
    unsigned short      uid;
    unsigned short      guid;
    unsigned int        mtime;
    unsigned int        inode_number;
    unsigned int        start_block;
    unsigned int        fragment;
    unsigned int        offset;
    unsigned int        file_size;
    unsigned int        block_list[0];
};

struct squashfs_lreg_inode_header {
    unsigned short      inode_type;
    unsigned short      mode;
    unsigned short      uid;
    unsigned short      guid;
    unsigned int        mtime;
    unsigned int        inode_number;
    squashfs_block      start_block;
    long long           file_size;
    long long           sparse;
    unsigned int        nlink;
    unsigned int        fragment;
    unsigned int        offset;
    unsigned int        xattr;
    unsigned int        block_list[0];
};

struct squashfs_dir_inode_header {
    unsigned short      inode_type;
    unsigned short      mode;
    unsigned short      uid;
    unsigned short      guid;
    unsigned int        mtime;
    unsigned int        inode_number;
    unsigned int        start_block;
    unsigned int        nlink;
    unsigned short      file_size;
    unsigned short      offset;
    unsigned int        parent_inode;
};

struct squashfs_ldir_inode_header {
    unsigned short      inode_type;
    unsigned short      mode;
    unsigned short      uid;
    unsigned short      guid;
    unsigned int        mtime;
    unsigned int        inode_number;
    unsigned int        nlink;
    unsigned int        file_size;
    unsigned int        start_block;
    unsigned int        parent_inode;
    unsigned short      i_count;
    unsigned short      offset;
    unsigned int        xattr;
    struct squashfs_dir_index   index[0];
};

struct squashfs_dir_entry {
    unsigned short      offset;
    short               inode_number;
    unsigned short      type;
    unsigned short      size;
    char                name[0];
};

struct squashfs_dir_header {
    unsigned int        count;
    unsigned int        start_block;
    unsigned int        inode_number;
};

struct squashfs_fragment_entry {
    long long           start_block;
    unsigned int        size;
    unsigned int        unused;
};

struct squashfs_xattr_entry {
    unsigned short      type;
    unsigned short      size;
};

struct squashfs_xattr_val {
    unsigned int        vsize;
};

struct squashfs_xattr_id {
    long long           xattr;
    unsigned int        count;
    unsigned int        size;
};

struct squashfs_xattr_table {
    long long           xattr_table_start;
    unsigned int        xattr_ids;
    unsigned int        unused;
};

// Compression options

struct gzip_comp_opts {
    int                 compression_level;
    short               window_size;
    short               strategy;
};

struct lzo_comp_opts {
    int                 algorithm;
    int                 compression_level;
};

struct xz_comp_opts {
    int                 dictionary_size;
    int                 flags;
};

struct lz4_comp_opts {
    int                 version;
    int                 flags;
};

struct zstd_comp_opts {
    int                 compression_level;
};
"""

c_squashfs = cstruct().load(squashfs_def)

INODE_STRUCT_MAP = {
    c_squashfs.SQUASHFS_DIR_TYPE: c_squashfs.squashfs_dir_inode_header,
    c_squashfs.SQUASHFS_FILE_TYPE: c_squashfs.squashfs_reg_inode_header,
    c_squashfs.SQUASHFS_SYMLINK_TYPE: c_squashfs.squashfs_symlink_inode_header,
    c_squashfs.SQUASHFS_BLKDEV_TYPE: c_squashfs.squashfs_dev_inode_header,
    c_squashfs.SQUASHFS_CHRDEV_TYPE: c_squashfs.squashfs_dev_inode_header,
    c_squashfs.SQUASHFS_FIFO_TYPE: c_squashfs.squashfs_base_inode_header,
    c_squashfs.SQUASHFS_SOCKET_TYPE: c_squashfs.squashfs_base_inode_header,
    c_squashfs.SQUASHFS_LDIR_TYPE: c_squashfs.squashfs_ldir_inode_header,
    c_squashfs.SQUASHFS_LREG_TYPE: c_squashfs.squashfs_lreg_inode_header,
    c_squashfs.SQUASHFS_LSYMLINK_TYPE: c_squashfs.squashfs_symlink_inode_header,
    c_squashfs.SQUASHFS_LBLKDEV_TYPE: c_squashfs.squashfs_ldev_inode_header,
    c_squashfs.SQUASHFS_LCHRDEV_TYPE: c_squashfs.squashfs_ldev_inode_header,
    c_squashfs.SQUASHFS_LFIFO_TYPE: c_squashfs.squashfs_lipc_inode_header,
    c_squashfs.SQUASHFS_LSOCKET_TYPE: c_squashfs.squashfs_lipc_inode_header,
}

TYPE_MAP = [
    None,
    stat.S_IFDIR,
    stat.S_IFREG,
    stat.S_IFLNK,
    stat.S_IFBLK,
    stat.S_IFCHR,
    stat.S_IFIFO,
    stat.S_IFSOCK,
    stat.S_IFDIR,
    stat.S_IFREG,
    stat.S_IFLNK,
    stat.S_IFBLK,
    stat.S_IFCHR,
    stat.S_IFIFO,
    stat.S_IFSOCK,
]



from dissect.squashfs.exceptions import (
    Error,
    FileNotFoundError,
    NotADirectoryError,
    NotAFileError,
    NotASymlinkError,
)
from dissect.squashfs.squashfs import FileStream, INode, SquashFS

__all__ = [
    "Error",
    "FileNotFoundError",
    "FileStream",
    "INode",
    "NotADirectoryError",
    "NotAFileError",
    "NotASymlinkError",
    "SquashFS",
]



from __future__ import annotations

import importlib

from dissect.squashfs.c_squashfs import c_squashfs


def initialize(id: int, options: bytes | None) -> Compression:
    # Options have no effect on decompression, so ignore for now
    modules = {
        c_squashfs.ZLIB_COMPRESSION: (NativeZlib,),
        c_squashfs.LZMA_COMPRESSION: (NativeLZMA,),
        c_squashfs.LZO_COMPRESSION: (AvailableLZO,),
        c_squashfs.XZ_COMPRESSION: (NativeXZ,),
        c_squashfs.LZ4_COMPRESSION: (AvailableLZ4,),
        c_squashfs.ZSTD_COMPRESSION: (NativeZSTD,),
    }

    try:
        for mod in modules[id]:
            try:
                return mod()
            except ModuleNotFoundError:  # noqa: PERF203
                pass
        else:
            raise ImportError(f"No modules available ({modules[id]})")  # noqa: TRY301
    except ImportError:
        raise ValueError(f"Compression ID {id} requested but module ({modules[id]}) is not available")
    except KeyError:
        raise NotImplementedError(f"Unsupported compression ID: {id}")


class Compression:
    module = None

    def __init__(self):
        self._module = importlib.import_module(self.module)

    def compress(self, data: bytes) -> bytes:
        raise NotImplementedError

    def decompress(self, data: bytes, expected: int) -> bytes:
        raise NotImplementedError


class NativeZlib(Compression):
    module = "zlib"

    def decompress(self, data: bytes, expected: int) -> bytes:
        return self._module.decompress(data)


class NativeLZMA(Compression):
    module = "lzma"

    def decompress(self, data: bytes, expected: int) -> bytes:
        # LZMA seems to behave inconsistently, even on the same version on different machines
        # Errors may occur here and there's unfortunately nothing we can do about it
        return self._module.decompress(data)


class AvailableLZO(Compression):
    module = "dissect.util.compression.lzo"

    def decompress(self, data: bytes, expected: int) -> bytes:
        return self._module.decompress(data, False, expected)


class NativeXZ(Compression):
    module = "lzma"

    def decompress(self, data: bytes, expected: int) -> bytes:
        return self._module.decompress(data)


class AvailableLZ4(Compression):
    module = "dissect.util.compression.lz4"

    def decompress(self, data: bytes, expected: int) -> bytes:
        return self._module.decompress(data, uncompressed_size=expected)


class NativeZSTD(Compression):
    module = "zstandard"

    def decompress(self, data: bytes, expected: int) -> bytes:
        return self._module.decompress(data)



# References:
# - https://dr-emann.github.io/squashfs/
# - Linux kernel

from __future__ import annotations

import stat
import struct
from bisect import bisect_right
from functools import cache, cached_property, lru_cache
from typing import TYPE_CHECKING, BinaryIO

from dissect.util import ts
from dissect.util.stream import RunlistStream

from dissect.squashfs import compression
from dissect.squashfs.c_squashfs import INODE_STRUCT_MAP, TYPE_MAP, c_squashfs
from dissect.squashfs.exceptions import (
    FileNotFoundError,
    NotADirectoryError,
    NotAFileError,
    NotASymlinkError,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime


class SquashFS:
    def __init__(self, fh: BinaryIO):
        self.fh = fh

        sb = c_squashfs.squashfs_super_block(fh)
        if sb.s_magic != c_squashfs.SQUASHFS_MAGIC:
            if sb.s_magic == c_squashfs.SQUASHFS_MAGIC_SWAP:
                raise NotImplementedError("Unsupported squashfs pre-4.0 big-endian filesystem")

            raise ValueError("Invalid squashfs superblock")

        # Only support squashfs 4.x for now
        if sb.s_major != 4:
            raise NotImplementedError("Unsupported squashfs version")

        self.sb = sb
        self.mkfs_time = ts.from_unix(self.sb.mkfs_time)
        self.block_size = self.sb.block_size
        self.block_log = self.sb.block_log
        self.flags = self.sb.flags
        self.major = self.sb.s_major
        self.minor = self.sb.s_minor
        self.size = self.sb.bytes_used

        self._read_block = lru_cache(1024)(self._read_block)
        self._lookup_id = lru_cache(1024)(self._lookup_id)
        self._lookup_inode = lru_cache(1024)(self._lookup_inode)
        self._lookup_fragment = lru_cache(1024)(self._lookup_fragment)

        self._compression_options = None
        if (self.sb.flags >> c_squashfs.SQUASHFS_COMP_OPT) & 1:
            self._compression_options = self._read_block(len(self.sb))[1]

        self._compression = compression.initialize(self.sb.compression, self._compression_options)

        self.id_table = self._read_table(self.sb.id_table_start, self.sb.no_ids, 4)
        self.lookup_table = self._read_table(self.sb.lookup_table_start, self.sb.inodes, 8)
        self.fragment_table = self._read_table(
            self.sb.fragment_table_start, self.sb.fragments, len(c_squashfs.squashfs_fragment_entry)
        )

        self.root = self.inode(self.sb.root_inode >> 16, self.sb.root_inode & 0xFFFF, name="/")

    def inode(
        self,
        block: int,
        offset: int,
        name: str | None = None,
        type: int | None = None,
        inode_number: int | None = None,
        parent: INode | None = None,
    ) -> INode:
        # squashfs inode numbers consist of a block number and offset in that block
        return INode(self, block, offset, name, type, inode_number, parent)

    def get(self, path: str | int, node: INode | None = None) -> INode:
        if isinstance(path, int):
            return self.inode(path >> 16, path & 0xFFFF)

        node = node or self.root

        parts = path.split("/")

        for part_num, part in enumerate(parts):
            if not part:
                continue

            if part == ".":
                continue

            if part == "..":
                node = node.parent or node
                continue

            while node.is_symlink() and part_num < len(parts):
                node = node.link_inode

            for entry in node.iterdir():
                if entry.name == part:
                    node = entry
                    break
            else:
                raise FileNotFoundError(f"File not found: {path}")

        return node

    def _read_table(self, offset: int, count: int, entry_size: int) -> list[int]:
        if not count or offset == c_squashfs.SQUASHFS_INVALID_BLK:
            return []

        num_bytes = count * entry_size
        num_blocks = (num_bytes + c_squashfs.SQUASHFS_METADATA_SIZE - 1) // c_squashfs.SQUASHFS_METADATA_SIZE

        self.fh.seek(offset)
        return c_squashfs.uint64[num_blocks](self.fh)

    def _read_metadata(self, block: int, offset: int, length: int) -> tuple[int, int, bytes]:
        result = []
        while length:
            next_block, data = self._read_block(block)
            remaining = len(data) - offset

            if remaining <= length:
                result.append(data[offset:])
                length -= remaining
                block = next_block
                offset = 0
            else:
                result.append(data[offset : offset + length])
                offset += length
                break

        return block, offset, b"".join(result)

    def _read_block(self, block: int, length: int | None = None) -> tuple[int, bytes]:
        if length is not None:
            # Data block
            compressed = length & c_squashfs.SQUASHFS_COMPRESSED_BIT_BLOCK == 0
            length = length & ~c_squashfs.SQUASHFS_COMPRESSED_BIT_BLOCK
        else:
            # Metadata block
            self.fh.seek(block)
            length = c_squashfs.uint16(self.fh)

            compressed = length & c_squashfs.SQUASHFS_COMPRESSED_BIT == 0
            length = length & ~c_squashfs.SQUASHFS_COMPRESSED_BIT
            block += 2

        self.fh.seek(block)
        data = self.fh.read(length)

        if not compressed:
            return block + length, data

        if self._compression is None:
            raise RuntimeError(f"Tried to read compressed block {block} but no compression initialized")

        return block + length, self._compression.decompress(data, self.block_size)

    def _read_fragment(self, fragment: int, offset: int, length: int) -> bytes:
        entry = self._lookup_fragment(fragment)
        _, data = self._read_block(entry.start_block, entry.size)
        return data[offset : offset + length]

    def _lookup_id(self, id: int) -> int:
        block, offset = divmod(id * 4, c_squashfs.SQUASHFS_METADATA_SIZE)
        _, _, data = self._read_metadata(self.id_table[block], offset, 4)
        return struct.unpack("<I", data)[0]

    def _lookup_inode(self, inode_number: int) -> INode:
        if inode_number <= 0 or inode_number > self.sb.inodes:
            raise IndexError(f"inode number out of bounds (1, {self.sb.inodes}): {inode_number}")
        block, offset = divmod((inode_number - 1) * 8, c_squashfs.SQUASHFS_METADATA_SIZE)
        _, _, data = self._read_metadata(self.lookup_table[block], offset, 8)
        return self.get(struct.unpack("<Q", data)[0])

    def _lookup_fragment(self, fragment: int) -> c_squashfs.squashfs_fragment_entry:
        fragment_offset = fragment * len(c_squashfs.squashfs_fragment_entry)
        block, offset = divmod(fragment_offset, c_squashfs.SQUASHFS_METADATA_SIZE)

        _, _, data = self._read_metadata(self.fragment_table[block], offset, len(c_squashfs.squashfs_fragment_entry))
        return c_squashfs.squashfs_fragment_entry(data)

    def iter_inodes(self) -> Iterator[INode]:
        for inum in range(1, self.sb.inodes + 1):
            yield self._lookup_inode(inum)


class INode:
    def __init__(
        self,
        fs: SquashFS,
        block: int,
        offset: int,
        name: str | None = None,
        type: int | None = None,
        inode_number: int | None = None,
        parent: INode | None = None,
    ):
        self.fs = fs
        self.block = block
        self.offset = offset
        self.name = name
        self._type = type
        self._inode_number = inode_number
        self.parent = parent

        self.listdir = cache(self.listdir)

    def __repr__(self) -> str:
        return f"<inode {self.inode_number} ({self.block}, {self.offset})>"

    def _metadata(
        self,
    ) -> tuple[
        c_squashfs.squashfs_base_inode_header
        | c_squashfs.squashfs_dir_inode_header
        | c_squashfs.squashfs_reg_inode_header
        | c_squashfs.squashfs_symlink_inode_header
        | c_squashfs.squashfs_dev_inode_header
        | c_squashfs.squashfs_ldir_inode_header
        | c_squashfs.squashfs_lreg_inode_header
        | c_squashfs.squashfs_ldev_inode_header
        | c_squashfs.squashfs_lipc_inode_header,
        int,
        int,
    ]:
        base_struct = c_squashfs.squashfs_base_inode_header

        block = self.fs.sb.inode_table_start + self.block
        data_block, data_offset, data = self.fs._read_metadata(block, self.offset, len(base_struct))

        header = base_struct(data)
        actual_struct = INODE_STRUCT_MAP.get(header.inode_type)

        if len(actual_struct) != len(base_struct):
            data_block, data_offset, data = self.fs._read_metadata(block, self.offset, len(actual_struct))

        if actual_struct != base_struct:
            header = actual_struct(data)

        self.header = header
        self.data_block = data_block
        self.data_offset = data_offset

        return header, data_block, data_offset

    @cached_property
    def header(
        self,
    ) -> (
        c_squashfs.squashfs_base_inode_header
        | c_squashfs.squashfs_dir_inode_header
        | c_squashfs.squashfs_reg_inode_header
        | c_squashfs.squashfs_symlink_inode_header
        | c_squashfs.squashfs_dev_inode_header
        | c_squashfs.squashfs_ldir_inode_header
        | c_squashfs.squashfs_lreg_inode_header
        | c_squashfs.squashfs_ldev_inode_header
        | c_squashfs.squashfs_lipc_inode_header
    ):
        header, _, _ = self._metadata()
        return header

    @cached_property
    def data_block(self) -> int:
        _, data_block, _ = self._metadata()
        return data_block

    @cached_property
    def data_offset(self) -> int:
        _, _, data_offset = self._metadata()
        return data_offset

    @property
    def inode_number(self) -> int:
        return self._inode_number or self.header.inode_number

    @property
    def type(self) -> int:
        return TYPE_MAP[self._type or self.header.inode_type]

    @property
    def mode(self) -> int:
        return self.header.mode | self.type

    @property
    def uid(self) -> int:
        return self.fs._lookup_id(self.header.uid)

    @property
    def guid(self) -> int:
        return self.fs._lookup_id(self.header.guid)

    @property
    def gid(self) -> int:
        return self.guid

    @property
    def mtime(self) -> datetime:
        return ts.from_unix(self.header.mtime)

    @property
    def size(self) -> int | None:
        if self.is_dir() or self.is_file():
            return self.header.file_size
        if self.is_symlink():
            return self.header.symlink_size
        return None

    def is_dir(self) -> bool:
        return self.type == stat.S_IFDIR

    def is_file(self) -> bool:
        return self.type == stat.S_IFREG

    def is_symlink(self) -> bool:
        return self.type == stat.S_IFLNK

    def is_block_device(self) -> bool:
        return self.type == stat.S_IFBLK

    def is_character_device(self) -> bool:
        return self.type == stat.S_IFCHR

    def is_device(self) -> bool:
        return self.is_block_device() or self.is_character_device()

    def is_fifo(self) -> bool:
        return self.type == stat.S_IFIFO

    def is_socket(self) -> bool:
        return self.type == stat.S_IFSOCK

    def is_ipc(self) -> bool:
        return self.is_fifo() or self.is_socket()

    @cached_property
    def link(self) -> str:
        if not self.is_symlink():
            raise NotASymlinkError(f"{self!r} is not a symlink")

        _, _, data = self.fs._read_metadata(
            self.data_block,
            self.data_offset,
            self.header.symlink_size,
        )
        return data.decode(errors="surrogateescape")

    @cached_property
    def link_inode(self) -> INode:
        link = self.link
        relnode = None if link.startswith("/") else self.parent
        return self.fs.get(self.link, relnode)

    def listdir(self) -> dict[str, INode]:
        return {inode.name: inode for inode in self.iterdir()}

    def iterdir(self) -> Iterator[INode]:
        if not self.is_dir():
            raise NotADirectoryError(f"{self!r} is not a directory")

        if self.size == 3:
            return

        start = self.fs.sb.directory_table_start + self.header.start_block
        offset = self.header.offset

        bytes_read = 0
        while bytes_read < self.size - 3:
            start, offset, data = self.fs._read_metadata(start, offset, len(c_squashfs.squashfs_dir_header))
            dir_header = c_squashfs.squashfs_dir_header(data)
            bytes_read += len(data)

            for _ in range(dir_header.count + 1):
                start, offset, data = self.fs._read_metadata(start, offset, len(c_squashfs.squashfs_dir_entry))
                dir_entry = c_squashfs.squashfs_dir_entry(data)
                bytes_read += len(data)

                start, offset, data = self.fs._read_metadata(start, offset, dir_entry.size + 1)
                bytes_read += len(data)

                yield self.fs.inode(
                    dir_header.start_block,
                    dir_entry.offset,
                    data.decode(errors="surrogateescape"),
                    dir_entry.type,
                    dir_header.inode_number + dir_entry.inode_number,
                    parent=self,
                )

    @cached_property
    def block_list(self) -> list[tuple[int | None, int]]:
        fragment = self.header.fragment
        file_size = self.header.file_size
        if fragment == c_squashfs.SQUASHFS_INVALID_FRAG:
            frag_bytes = 0
            blocks = (file_size + self.fs.block_size - 1) >> self.fs.block_log
        else:
            blocks, frag_bytes = divmod(file_size, self.fs.block_size)

        if blocks:
            _, _, data = self.fs._read_metadata(
                self.data_block,
                self.data_offset,
                blocks * 4,
            )
            block_list = [(block, 1) for block in c_squashfs.uint32[blocks](data)]
        else:
            block_list = []

        if frag_bytes:
            block_list.append((None, frag_bytes))

        return block_list

    def open(self) -> FileStream:
        if not self.is_file():
            raise NotAFileError(f"{self!r} is not a file")

        return FileStream(self)


class FileStream(RunlistStream):
    def __init__(self, inode: INode):
        super().__init__(inode.fs.fh, inode.block_list, inode.size, inode.fs.block_size)
        self.inode = inode
        self.fs = inode.fs
        self.start_block = inode.header.start_block
        self.fragment = inode.header.fragment
        self.fragment_offset = inode.header.offset

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        block_offset = offset // self.block_size

        run_idx = bisect_right(self._runlist_offsets, block_offset)
        runlist_len = len(self.runlist)
        size = self.size

        start_block = self.start_block + sum(
            v & ~c_squashfs.SQUASHFS_COMPRESSED_BIT_BLOCK for v, _ in self.runlist[:run_idx]
        )

        while length > 0:
            if run_idx >= runlist_len:
                # We somehow requested more data than we have runs for
                break

            run_block_size, run_block_count = self.runlist[run_idx]

            # For squashfs we use 0 to indicate a sparse block and None to indicate a fragment
            if run_block_size is None:
                result.append(self.fs._read_fragment(self.fragment, self.fragment_offset, run_block_count))
                offset += run_block_count
                length -= run_block_count
            else:
                run_pos = offset - run_idx * self.block_size
                run_remaining = self.block_size - run_pos

                # Sometimes the self.size is way larger than what we actually have runs for?
                # Stop reading if we reach a negative run_remaining
                if run_remaining < 0:
                    break

                read_count = min(size - offset, min(run_remaining, length))

                # Sparse run
                if run_block_size == 0:
                    result.append(b"\x00" * read_count)
                else:
                    start_block, data = self.fs._read_block(start_block, run_block_size)
                    result.append(data[:read_count])

                offset += read_count
                length -= read_count
                run_idx += 1

        return b"".join(result)
